package jagafa;

import jagafa.ai.ProbabilityAI;
import jagafa.ai.core.JassAI;
import jagafa.flags.TestingFlags;
import jagafa.object.Card;
import jagafa.object.Player;
import jagafa.rule.RuleSet;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.Timer;

public class JassRound extends JassRoundCore {

	/**
	 * Constructor: init the game with the ruleSet given
	 */
	public JassRound(RuleSet rules) {
		super(rules);

	}

	/**
	 * Initialize all the players
	 * TODO: initPlayers() should be externalized or at least parameterized
	 */
	protected void initPlayers() {
		Player.resetPlayerCount();
		for (int i = 0; i < 4; i++) {
			Player newPlayer = null;
			/* Add human or ai-driven player */
			if (TestingFlags.playerIsHuman_[i]) {
				newPlayer = new Player(9, true);

			} else {
				newPlayer = new Player(9, false, new ProbabilityAI());
				newPlayer.getAI().init(newPlayer, this);
			}
			addPlayer(newPlayer);

		}
	}

	public void start(Player startPlayer) {
		if (this.running_) {
			return;
		}
		this.setLastTurnWinner(startPlayer);
		this.setActivePlayer(startPlayer);
		start();
	}

	/**
	 * Start the game by installing the main and refresh timers
	 * Only if the RuleSet is set!
	 */
	public void start() {
		
		if (this.running_) {
			return;
		}
		
		if (this.getRules() != null) {
			
			this.running_ = true;
			installTimer();
		}
	}

	/**
	 * Adds a card of Player p to the active Board, if valid
	 */
	public synchronized void playCard(Card card, Player player) {

		this.activeBoard().addCard(card, player);

		this.nextPlayer();

		/* 
		 * If four cards are on the board, update scores, cleanup and refresh 
		 * Otherwise do some AI processing
		 */
		if (this.activeBoard().cardsPlayed() == 4) {

			Player winner = getWinner(this.activeBoard());
			this.getGoneHeap().addTurn(this.activeBoard(), winner);
			winner.addScore(this.getScore(this.activeBoard()));

			/* if round is over, add lastRound score, stop the ai timer and report 
			 * scores to the top-level score table if there is any 
			 */
			if (isRoundOver()) {
				winner.addScore(5);

				reportScores();
				aiTimer_.stop();
				
			} else {
				this.getRules().updateAfterTurn();
			}

			/* Set the active Player to the winner */
			this.setLastTurnWinner(winner);
			this.setActivePlayer(winner);

			/* Start the refresh timer */
			refreshTimer_.start();
		} else {

			aiTimer_.start();
			this.toggleNotify();
		}

	}

	/**
	 * Do all computer movements
	 */
	public synchronized void doComputer() {

		if (activePlayer().isComputer()) {
			JassAI ai = activePlayer().getAI();
			ai.init(activePlayer(), this);
			activePlayer().playCard(ai.computeCard(), this);
			toggleNotify();

		}
		/* If the game is over, repeat the ai timer only once to cleanup */
		if (this.isRoundOver()) {
			aiTimer_.setRepeats(false);
		}

	}

	/**
	 * Print the score to stdout
	 */
	public void printScores() {
		int sum1 = this.getPlayer(0).getScore() + this.getPlayer(2).getScore();
		int sum2 = this.getPlayer(1).getScore() + this.getPlayer(3).getScore();
		System.out.println("Player 0 and 2: " + sum1);
		System.out.println("Player 1 and 3: " + sum2);
	}

	/**
	 * Install the ai- and refresh-timers
	 */
	private void installTimer() {
		/**
		 * The refresh timer updates the ui every 1.5 seconds
		 * In addition he clears the board if a turn/round is over
		 */
		int uiDelay = TestingFlags.UI_REFRESH_DELAY;
		if (TestingFlags.speedUp_) {
			uiDelay = TestingFlags.FAST_UI_REFRESH_DELAY;
		}
		refreshTimer_ = new Timer(uiDelay, new ActionListener() {

			public void actionPerformed(ActionEvent e) {

				if (activePlayer() == getLastTurnWinner()) {
					activeBoard().clearBoard();
					toggleNotify();
				}

			}

		});
		refreshTimer_.setRepeats(false);

		/**
		 * The AI timer does a computer move and restarts the game if needed
		 */
		int aiDelay = TestingFlags.NORMAL_AI_DELAY;
		if (TestingFlags.speedUp_) {
			aiDelay = TestingFlags.FAST_AI_DELAY;
		}

		aiTimer_ = new Timer(aiDelay, new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				if (TestingFlags.paused_) {
					return;
				}
				if (!refreshTimer_.isRunning()) {
					//for (int i = 0; i < 4; i++) {
					doComputer();
					//	}
				}
				if (isRoundOver()) {
					running_ = false;
					//restartTimer_.start();
					//restartTimer_.setRepeats(false);
				}

			}

		});
		aiTimer_.setRepeats(true);
		aiTimer_.start();
	}
}
