package jagafa.ai.core;

import jagafa.JassRound;
import jagafa.flags.AIFlags;
import jagafa.naming.ESwissColorNames;
import jagafa.object.Board;
import jagafa.object.Card;
import jagafa.object.GoneCardHeap;
import jagafa.object.Hand;
import jagafa.object.Player;
import jagafa.stats.PlayerStats;
import jagafa.stats.RoundStats;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * ProbabylityTable provides functions to calculate probabilities
 * for all players having cards of a color or having a specific card (all 
 * from the view of a Player)
 * ProbabilityTable is the core of the ProbabilityAI AI implementation
 * @see class ProbabilityAI
 */
public class ProbabilityTable {

	private static final int DECK_SIZE = 36;

	private static final int MAX_HAND_SIZE = 9;

	private static final int PLAYERS = 4;

	private static final int COLORS = 4;

	private JassRound round_;

	private Player viewPlayer_;

	private List<Map<Card, Float>> cardProb_;

	/**
	 * colorProb_: 
	 * ------------------------------------------------ 
	 * Stores the Color probabilities for each player 
	 * % Map playerColorMap = colorProb_.get(playerNr); 
	 * returns the probability map for player playerNr
	 */
	private List<Map<Integer, Float>> colorProb_;

	//private GoneCardHeap goneCards_;

	private int playerNr;

	//	private int nthThrowColor[][];
	private List<List<Integer>> throwColors_;

	private boolean[][] canHaveColor;

	private int[] othersAvailableOfColor_;

	private int[] ownCardsOfColor;

	private int[] goneOfColor_;

	/**
	 * Constructor: Create the ProbabilityTable and init all values
	 * @param p The viewpoint Player
	 * @param game The game object
	 */
	public ProbabilityTable(Player p, JassRound game) {
		round_ = game;
		viewPlayer_ = p;

		playerNr = round_.getPlayerNr(p);

		//goneCards_ = round_.getGoneHeap();

		initMaps();

		update();
	}

	/**
	 * Update the table
	 */
	public void update() {
		RoundStats roundStats = new RoundStats(this.round_);
		PlayerStats playerStats = new PlayerStats(this.round_, viewPlayer_);

		Hand playerHand = this.viewPlayer_.getHand();
		/* goneOfColor[colorNr] holds the number of Cards 
		 * in the GoneCardHeap of Color colorNr */
		goneOfColor_ = new int[COLORS];

		/* handSize[playerNr] holds the number of Cards 
		 * in the Hand of Player playerNr */
		int handSize[] = new int[PLAYERS];

		/* cardsOfColorOwn[colorNr] holds number of Card 
		 * of Color colorNr in viewPlayers Hand */
		ownCardsOfColor = new int[COLORS];

		//nthThrowColor = new int[PLAYERS][MAX_HAND_SIZE];
		throwColors_ = new LinkedList<List<Integer>>();

		for (int ply = 0; ply < PLAYERS; ply++) {
			throwColors_.add(getThrowColorList(this.round_.getPlayer(ply)));
			handSize[ply] = this.round_.getPlayer(ply).getHand().size();

		}

		/* Init some Basic calculation-helper variables */

		for (int i = 0; i < COLORS; i++) {
			goneOfColor_[i] = roundStats.getGoneList()[i].size();

			ownCardsOfColor[i] = playerStats.getColorList()[i].size();
		}

		/*
		 * Cards left in the game without OWN
		 */
		cardsWithoutOwn_ = DECK_SIZE - this.round_.getGoneHeap().size()
				- handSize[this.playerNr];

		/*
		 * Wieviele Karten von color sind noch im Spiel abgesehen von OWN
		 */
		othersAvailableOfColor_ = new int[COLORS];
		for (int color = 0; color < COLORS; color++) {
			othersAvailableOfColor_[color] = MAX_HAND_SIZE - ownCardsOfColor[color]
					- goneOfColor_[color];
		}

		canHaveColor = getCanHaveColorArray(this.viewPlayer_);

		//System.out.println("View of player: " + this.player_.getName());

		createColorProb();
		createCardProb();
	}

	private int cardsWithoutOwn_;

	/**
	 * Returns a 2 dimensional array of boolean values.
	 * canHaveColor[playerNr][colorNr] is false, if there is no
	 * chance that player playerNr has any cards of the color colorNr
	 */
	private boolean[][] getCanHaveColorArray(Player viewPlayer) {
		int viewPlayerNr = this.round_.getPlayerNr(viewPlayer);

		boolean canHaveColor[][] = new boolean[PLAYERS][COLORS];
		for (int plaNr = 0; plaNr < PLAYERS; plaNr++) {
			for (int color = 0; color < COLORS; color++) {

				canHaveColor[plaNr][color] = hasAlwaysPlayedColor(this.round_
						.getPlayer(plaNr), color);

				if (othersAvailableOfColor_[color] == 0 || goneOfColor_[color] == 9) {
					if (plaNr != viewPlayerNr) {
						canHaveColor[plaNr][color] = false;
					}
				}
			}
		}
		return canHaveColor;
	}

	/**
	 * Create the ProbabilityTable's color probabilities
	 *
	 */
	private void createColorProb() {

		for (int ply = 0; ply < PLAYERS; ply++) {
			Map<Integer, Float> colorMap = new HashMap<Integer, Float>();

			for (int col = 0; col < COLORS; col++) {
				/* If it's the viewpoint Player: continue with next player */
				if (ply == playerNr) {
					continue;
				}

				float prob = 0;
				/* If the player has none of this color set prob to 0 */
				if (!this.canHaveColor[ply][col]) {
					prob = 0;

				} else {
					for (int k = 0; k < throwColors_.get(ply).size(); k++) {
						/*
						 * if the player has thrown away the color this reduces
						 * the prob that he has some/good left (only count once)
						 */
						if (col == throwColors_.get(ply).get(k).intValue()) {
							prob -= (throwColors_.get(ply).size() - k)
									* AIFlags.AI_THROWCOLOR_MULTIPLICATOR;
						}

						/*if (col == this.nthThrowColor[ply][k]) {
						 prob -= (MAX_HAND_SIZE - k) * AIFlags.AI_THROWCOLOR_MULTIPLICATOR;
						 break;
						 }*/

					}
					int othersOfColor = this.othersAvailableOfColor_[col];

					int othersCardsInGame = cardsWithoutOwn_;
					if (othersCardsInGame != 0) {
						prob += ((double) othersOfColor / (double) othersCardsInGame);
					} else {
						prob = 0;
					}
				}

				float roundedProb = ((float) Math.rint(prob * 1000.0f)) / 1000.0f;

				colorMap.put(new Integer(col), new Float(roundedProb));
			}

			this.colorProb_.add(colorMap);
		}

	}

	/**
	 * Create the ProbabilityTable's card probabilities
	 *
	 */
	private void createCardProb() {

	}

	/**
	 * Liefert true, wenn player bei color immer Farbe gespielt hat 
	 * false, falls player color einmal nicht Gefarbt hat (eine andere 
	 * Farbe verworfen hat)
	 */
	private boolean hasAlwaysPlayedColor(Player player, int color) {
		GoneCardHeap goneCards = this.round_.getGoneHeap();
		boolean colorPlayed = true;
		/* for all played turns */
		for (int i = 0; i < goneCards.turns(); i++) {
			Board bo = goneCards.getBoard(i);
			int firstColor = (int) bo.get(0).getColor();

			/* Ist diese Runde von der Farbe color (relevante Farbe)? */
			if (firstColor == color) {
				Card playedCard = goneCards.getPlayedCard(player, bo);

				/* Hat der spieler nicht gefarbt? */
				if (!(playedCard.getColor() == color)) {
					return false;
				}
			}

		}
		return true;
	}

	private List<Integer> getThrowColorList(Player player) {
		List<Integer> throwList = new LinkedList<Integer>();
		GoneCardHeap goneCards = this.round_.getGoneHeap();

		int nthThrow[] = new int[MAX_HAND_SIZE];

		for (int nth = 0; nth < MAX_HAND_SIZE; nth++) {
			nthThrow[nth] = -1;
		}

		/* For all played turns */
		for (int turn = 0; turn < goneCards.turns(); turn++) {
			Board bo = goneCards.getBoard(turn);
			int firstColor = bo.get(0).getColor();
			Card playedCard = goneCards.getPlayedCard(player, bo);

			/* Hat der Spieler hier nicht gefarbt? */
			if (!(playedCard.getColor() == firstColor)) {
				throwList.add(Integer.valueOf(playedCard.getColor()));
			}
		}

		return throwList;

		//return nthThrow;
	}

	private void initMaps() {
		colorProb_ = new LinkedList<Map<Integer, Float>>();
		cardProb_ = new LinkedList<Map<Card, Float>>();

		for (int i = 0; i < COLORS; i++) {
			TreeMap<Integer, Float> colprobi = new TreeMap<Integer, Float>();
		}
	}

	/**
	 * Get the probability of a Player having cards of a specific color
	 * 
	 * @param p The relevant Player
	 * @param color The relevant color
	 * @return A double representing the chance of player p having a card of color 
	 */
	public float getColorProb(Player p, int color) {
		int colorNr = round_.getPlayerNr(p);
		Map<Integer, Float> playerMap = colorProb_.get(colorNr);
		return Math.round(1000.0f*playerMap.get(new Integer(color)).floatValue())/1000.0f;
	}

	/**
	 * Get the probability of a Player having a specific card
	 * 
	 * @param p The relevant Player
	 * @param c The relevant card
	 * @return A double value representing the chance of player p having c in his hand
	 */
	public float getCardProb(Player p, Card c) {
		int playerNr = round_.getPlayerNr(p);
		Map<Card, Float> playerMap = cardProb_.get(playerNr);
		return playerMap.get(c).floatValue();
	}

	public String toString() {
		String res = new String();
		res += ("Color prob: Player " + this.playerNr + "'s view:\n");
		res += "Player: \t";
		for (int ply = 0; ply < 4; ply++) {
			if (ply != this.playerNr) {

				res += (ply + "\t");
			}
		}
		res += "\n";
		for (int col = 0; col < 4; col++) {
			res += (ESwissColorNames.getName(col)) + " =\t";
			for (int ply = 0; ply < 4; ply++) {
				if (ply != this.playerNr) {

					Float prob = this.getColorProb(this.round_.getPlayer(ply),
							new Integer(col));
					res += prob + "\t";

				}
				
			}
			res += "\n";
		}

		int ply = this.round_.getPartner(this.viewPlayer_).getID();
		res += ("\nPartner Throws: ");
		for (int i = 0; i < this.throwColors_.get(ply).size(); i++) {
			int color = this.throwColors_.get(ply).get(i).intValue();
			res += (i + ". " + ESwissColorNames.getName(color) + ", ");

		}

		res += "\n";

		return res;
	}
}
