/**
 * 
 */
package jagafa.view.newui;

import jagafa.JassRound;
import jagafa.flags.AIFlags;
import jagafa.flags.TestingFlags;
import jagafa.object.Player;
import jagafa.rule.RuleSet;
import jagafa.scores.ScoreTable;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Observable;
import java.util.Observer;

import javax.swing.Timer;

/**
 * RoundController2 collects UI actions from the view and sends it to the 
 * round object.
 * He also gets UI-update notification of the round and sends it to the view
 * 
 * This version of RoundController does not show anything on the screen. It just creates
 * the views. A top-level controller can fetch those views (JPanel) by calling the get...()
 * methods.
 */
public class RoundController2 extends Observable implements ActionListener, Observer {

	private static final int PLAYERS = 4;

	private JassRound round_;

	private JassRoundPanel roundPanel_;

	private RoundStatsPanel statView_;

	private ScoreTable scoreTable_;

	private RuleChoicePanel ruleChoice_;

	private SimpleSTPanel stPanel_;

	/** 
	 * Constructor does nothing. Game has to be started by calling
	 * 	 startGame(Rules r)
	 */
	public RoundController2() {
		installRestartTimer();

	}

	/** Update the ui according to changes in the round_ */
	private void updateUI() {
		
		for (int i = 0; i < PLAYERS; i++) {
			if (TestingFlags.showAllHands_) {
				roundPanel_.showHand(i);
			} else {

				if (this.round_.getPlayer(i).isComputer()) {
					if (!TestingFlags.showHand_[i]) {
						roundPanel_.hideHand(i);
					} else {
						roundPanel_.showHand(i);
					}
				} else {
					roundPanel_.showHand(i);
				}
			}
			roundPanel_.updateHandPanel(round_.getPlayer(i).getHand(), i);

		}

		roundPanel_.getBoardPanel().updateBoardUI(round_.activeBoard());

		roundPanel_.repaint();
		this.stPanel_.update();
		if (TestingFlags.showRoundStats_) {
			this.updateStats();
		}
		
		this.setChanged();
		this.notifyObservers();
		this.clearChanged();
		
	}

	/**
	 * Called by the top-level UI's cardPressed() method
	 */
	public void actionPerformed(ActionEvent e) {
		userAction(e);

	}

	/**
	 * Called when the User pushes a CardButton (through a ActionListener)
	 */
	private void userAction(ActionEvent e) {
		/* Extract hand and card-number */
		String message = e.getActionCommand();

		String hand = message.substring(0, 1);
		String card = message.substring(1);

		int playerNrOfHand = Integer.valueOf(hand).intValue();
		int cardNrOfHand = Integer.valueOf(card).intValue();

		/* check wether the hand is from the active player or not */
		if (round_.activePlayerNr() == playerNrOfHand) {
			if (!round_.activePlayer().isComputer()) {
				round_.activePlayer().playCard(cardNrOfHand, round_);
			}
		}

		/* update the UI */
		updateUI();
	}

	/**
	 * Restart the controller
	 * Should not be used though a controller should only be used for one round
	 * HINT: Create a new Controller for another round. 
	 * 		 Just reuse the same ScoreTable object for the new Controller
	 * 
	 * @deprecated
	 */

	/**
	 * Starts the game with the ruleSet given
	 */
	public void startRound(ScoreTable scoreTable) {
		this.scoreTable_ = scoreTable;

		/* create the UI */
		this.stPanel_ = new SimpleSTPanel(this.scoreTable_);
		roundPanel_ = new JassRoundPanel();

		/* Add the card Listeners which will call cardPressed */
		roundPanel_.addCardListener(this);

		/* create the game engine with no RuleSet
		 * The ruleset will be defined before calling start() with
		 * round_.setRules(RuleSet rule) 
		 * */
		round_ = new JassRound(null);
		round_.addObserver(this);

		round_.setScoreTable(this.scoreTable_);
		
		/* create stats UI if needed */
		if (TestingFlags.showRoundStats_) {
			statView_ = new RoundStatsPanel(round_);
			statView_.setRound(round_);
		}

		/* Init RuleChoice UI */
		ruleChoice_ = new RuleChoicePanel(round_, roundPanel_);

		/* update the UI */
		updateUI();
		if (TestingFlags.automatedTest_) {
			this.runAutomated();
		}
		this.stPanel_.update();
	}

	/* Update the stat view */
	public void updateStats() {
		statView_.update(this.round_.activePlayer());
		this.stPanel_.update();
	}

	/**
	 * Update the UI when an observable calls this observer
	 */
	public void update(Observable o, Object arg) {
		updateScoreTable();

		if (!TestingFlags.silentTest_) {
			updateUI();
		}

		/* If the round is over, try a restart */
		if (this.round_.isRoundOver()) {

			if (TestingFlags.repeatedTest_) {
				if (TestingFlags.maxRepeats_ > 0) {
					if (TestingFlags.repeats_ == TestingFlags.maxRepeats_) {
						TestingFlags.repeatedTest_ = false;
						System.out.println("Stopped after " + TestingFlags.repeats_
								+ " rounds");
						System.out.println("GradeWeight: " + AIFlags.GRADE_WEIGHT);

						System.out.println(scoreTable_);
						return;
					}
					TestingFlags.repeats_++;
				}

				restartTimer_.setRepeats(false);
				restartTimer_.start();

			}
		}

	}

	/**
	 * Does nothing yet
	 *
	 */
	private void updateScoreTable() {

	}

	/* The restart timer: 
	 * TODO: remove after a top-level GameController is introduced 
	 */
	private Timer restartTimer_;

	/**
	 * Install the restart timer
	 *
	 */
	private void installRestartTimer() {
		/**
		 * The restart timer is needed to wait for a new game to start
		 * If he is called, then the game waits 6 seconds, then the game 
		 * will reset and restart itself if the game is endless
		 */
		restartTimer_ = new Timer(TestingFlags.RESTART_DELAY, new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				if (TestingFlags.repeatedTest_) {
					round_.reset();
				
					if (round_.getPlayer(scoreTable_.getStartPlayer()).isComputer()) {
						runAutomated();
					} else {
						System.out.println("Current player is not a Computer player!");
					}
				}
			}

		});
		restartTimer_.setRepeats(false);
	}

	/**
	 * Get the round instance of the controller
	 */
	public JassRound getRound() {
		return this.round_;
	}

	/**
	 * Get the rounds UI panel
	 */
	public JassRoundPanel getRoundPanel() {
		return this.roundPanel_;
	}

	/**
	 * Get the stats panel
	 */
	public RoundStatsPanel getStatsPanel() {
		return this.statView_;
	}

	/**
	 * Get the RuleChoice panel
	 */
	public RuleChoicePanel getRuleChoicePanel() {
		return this.ruleChoice_;
	}

	public void runAutomated() {
		this.round_.reset();
		Player startPlayer = this.round_.getPlayer(this.scoreTable_.getStartPlayer());
		
		Player choosingPlayer = this.round_.getPlayer(this.scoreTable_.getChoosingPlayer());
		
		RuleSet choice = choosingPlayer.getAI().chooseRuleSet();
		if (choice == null) {
			this.scoreTable_.changeChoosingPlayer();
			choosingPlayer = this.round_.getPlayer(this.scoreTable_
					.getChoosingPlayer());
			startPlayer = this.round_
					.getPlayer(this.scoreTable_.getStartPlayer());

			choice = choosingPlayer.getAI().chooseRuleSet();
		}
		this.round_.setRules(choice);
		this.round_.start(startPlayer);
		this.ruleChoice_.updateRuleLabel();
		
		this.round_.toggleNotify();
	}

	
	public SimpleSTPanel getSTPanel() {
		return this.stPanel_;
	}

	public void toggleNoify() {
		this.setChanged();
		this.notifyObservers();
		this.clearChanged();
		
	}
}
